//
// Command line parser
//

//
// windows.h checks for _WIN32_WINNT. If this is defined and greater than 0x0400 then
// winsock2.h is included, otherwise winsock.h is included. Unless WIN32_MEAN_AND_LEAN
// is defined, in which case neither is included.
//

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0400
#else
#if _WIN32_WINNT<0x0400
#error Need Winsock2 features available with _WIN32_WINNT>=0x0400!
#endif
#endif

#include <winsock2.h>
#include <windows.h>
#include <tchar.h>
#include "cmdline.h"

#include <stdlib.h>
#include <stdio.h>

inline LPCTSTR arg_pre(const CmdLineArgs &arg)
{
    if (arg.longopt) return _T("--");
    return _T("-");
}

inline LPCTSTR arg_str(const CmdLineArgs &arg)
{
    if (arg.longopt) return arg.longopt;
    return arg.shortopt;
}

void CCmdLine::CommonConstruct()
{
    m_numopts = 0;
    m_appopt = -1;
    m_valid = false;
    m_errmsg = NULL;
    m_vals = NULL;

    m_noval.supplied = false;
    m_noval.val.strval = NULL;

    int i;

    for ( i=0 ; ; i++ ) {
        if (m_opts[i].type==cmdtApp) {
            m_appopt = i;
        }
        if (m_opts[i].type==cmdtEnd) {
            m_numopts = i;
            break;
        }
    }

    if (m_numopts==0) return;

    m_vals = new CmdLineVal[m_numopts];
    for ( i=0 ; i<m_numopts ; i++ ) m_vals[i] = m_noval;

    if (Parse()) m_valid = true;
}

CCmdLine::CCmdLine(const CmdLineArgs *opts, LPCTSTR cmdline)
{
    m_cmdline = _tcsdup(cmdline);
    m_argc = 0;
    m_argv = NULL;
    m_opts = opts;

    CommonConstruct();
}

CCmdLine::CCmdLine(const CmdLineArgs *opts, int argc, char *argv[])
{
    m_cmdline = NULL;
    m_argc = argc;
    m_argv = new char*[m_argc];
    for ( int i=0 ; i<argc ; i++ ) m_argv[i] = strdup(argv[i]);
    m_opts = opts;

    CommonConstruct();
}

CCmdLine::~CCmdLine()
{
    m_valid = false;

    if (m_vals) {
        for ( int i=0 ; i<m_numopts ; i++ ) {
            if (m_opts[i].type==cmdtString && m_vals[i].val.strval) {
                free(const_cast<LPTSTR>(m_vals[i].val.strval));
            }
        }
        delete[] m_vals;
    }
    m_vals = NULL;

    if (m_cmdline) free(const_cast<char*>(m_cmdline));
    m_cmdline = NULL;

    if (m_argv) {
        for ( int i=0 ; i<m_argc ; i++ ) free(m_argv[i]);
        delete[] m_argv;
    }
    m_argv = NULL;
}

bool CCmdLine::IsValid()
{
    return m_valid;
}

void CCmdLine::PrintUsage()
{
    if (m_appopt>=0) {
        _ftprintf(stderr, _T("%s: %s\n"), (m_argv && m_argc>0 && strlen(m_argv[0])>0) ? m_argv[0] : m_opts[m_appopt].shortopt, m_opts[m_appopt].desc);
    }
    if (m_errmsg) {
        _ftprintf(stderr, _T("%s\n"), m_errmsg);
    }

    _ftprintf(stderr, _T("\nUsage:\n"));

    for ( int i=0 ; i<m_numopts ; i++ ) {
        switch (m_opts[i].type) {
        case cmdtFlag:
        case cmdtString:
        case cmdtInt:
        case cmdtFloat:
        case cmdtIPV4:
            if (m_opts[i].shortopt && m_opts[i].longopt) {
                _ftprintf(stderr, _T("    -%s --%s:\t%s\n"), m_opts[i].shortopt, m_opts[i].longopt, m_opts[i].desc);
            } else if (m_opts[i].shortopt) {
                _ftprintf(stderr, _T("    -%s:\t%s\n"), m_opts[i].shortopt, m_opts[i].desc);
            } else if (m_opts[i].longopt) {
                _ftprintf(stderr, _T("    --%s:\t%s\n"), m_opts[i].longopt, m_opts[i].desc);
            }
            break;
        }
    }

    _ftprintf(stderr, _T("\n"));
}

int CCmdLine::GetArgIndex(LPCTSTR opt, int argt)
{
    for ( int i=0 ; i<m_numopts ; i++ ) {
        switch (m_opts[i].type) {
        case cmdtFlag:
        case cmdtString:
        case cmdtInt:
        case cmdtFloat:
        case cmdtIPV4:
            if ((m_opts[i].shortopt && (argt & 1) && _tcscmp(opt, m_opts[i].shortopt)==0) ||
                (m_opts[i].longopt && (argt & 2) && _tcscmp(opt, m_opts[i].longopt)==0)) return i;
        }
    }

    return -1;
}

bool CCmdLine::IsArgSupplied(LPCTSTR opt, bool nodefault)
{
    if (!m_valid) return false;

    int i = GetArgIndex(opt);
    if (i<0) return false;

    return m_vals[i].supplied || (!nodefault && m_opts[i].defval);
}

CmdLineArg& CCmdLine::GetArg(LPCTSTR opt)
{
    if (!m_valid) return m_noval.val;

    int i = GetArgIndex(opt);
    if (i<0) return m_noval.val;

    return m_vals[i].val;
}

bool CCmdLine::Parse()
{
    if (m_cmdline) return ParseCmdLine();
    if (m_argv) return ParseArgv();

    return false;
}

void CCmdLine::SetErrorMsg(LPCTSTR fmt, ...)
{
    va_list args;
    va_start(args, fmt);

    TCHAR buf[512];

    _vsntprintf(buf, 512, fmt, args);
    if (m_errmsg) free(m_errmsg);
    m_errmsg = _tcsdup(buf);

    va_end(args);
}


bool CCmdLine::ParseCmdLine()
{
    return false;
}

bool CCmdLine::ParseArgv()
{
    int i = 1;
    while (i<m_argc) {
        if (m_argv[i][0]=='-' || m_argv[i][1]=='/') {
            int j = -1;
            if (m_argv[i][0]=='-' && m_argv[i][1]=='-') {
                j = GetArgIndex(m_argv[i]+2, 2);
            } else {
                j = GetArgIndex(m_argv[i]+1, 1);
            }

            if (j<0) {
                SetErrorMsg(_T("Unknown option %s"), m_argv[i]);
                return false;
            }

            if (m_opts[j].type==cmdtFlag) {
                m_vals[j].supplied = true;
                m_vals[j].val.flagval = true;
                i++;
            } else if (i+1==m_argc) {
                SetErrorMsg(_T("Missing parameter to %s option"), m_argv[i]);
                return false;
            } else {
                if (!ParseVal(j, m_argv[i+1])) return false;
                m_vals[j].supplied = true;
                i += 2;
            }
        } else {
            for ( int j=0 ; j<m_numopts ; j++ ) {
                if ((m_opts[j].flags & cmdfPositional) && !m_vals[j].supplied)  {
                    if (!ParseVal(j, m_argv[i])) return false;
                    m_vals[j].supplied = true;
                    break;
                } else if (j+1==m_numopts) {
                    SetErrorMsg(_T("Too many parameters"));
                    return false;
                }
            }

            i++;
        }
    }

    for ( i=0 ; i<m_numopts ; i++ ) {
        if (!m_vals[i].supplied) {
            if (m_opts[i].defval) {
                if (!ParseVal(i, m_opts[i].defval)) {
                    SetErrorMsg(_T("Internal command line processing error (%d)"), i);
                    return false;
                }
            } else if (m_opts[i].flags & cmdfMandatory) {
                SetErrorMsg(_T("Missing required parameter %s%s"), arg_pre(m_opts[i]), arg_str(m_opts[i]));
                return false;
            }
        }
    }

    return true;
}

bool CCmdLine::ParseVal(int i, LPCTSTR val)
{
    switch (m_opts[i].type) {
    case cmdtString:
        m_vals[i].val.strval = _tcsdup(val);
        return true;
    case cmdtInt:
        if (_stscanf(val, _T("%i"), &m_vals[i].val.intval)!=1) {
            SetErrorMsg(_T("Invalid integer parameter %s"), val);
            return false;
        }
        return true;
    case cmdtFloat:
        if (_stscanf(val, _T("%lf"), &m_vals[i].val.floatval)!=1) {
            SetErrorMsg(_T("Invalid numeric parameter %s"), val);
            return false;
        }
        return true;
    case cmdtIPV4:
        m_vals[i].val.inetval = ntohl(inet_addr(val));
        if (m_vals[i].val.inetval==INADDR_NONE) {
            SetErrorMsg(_T("Invalid IP address parameter %s"), val);
            return false;
        }
        return true;
    }
    return false;
}